from great_expectations.datasource import PandasDatasource


class MyCustomV2ApiDatasource(PandasDatasource):
    """
    This class is used only for testing.
    E.g. ensuring appropriate usage stats messaging when using plugin functionality.
    """

    pass
