{% if batch_request %}
validator.{{ expectation["expectation_type"] }}({{ kwargs_string }}{{ meta_args }})
{% else %}
expectation_configuration = ExpectationConfiguration(**{{ expectation }})
suite.add_expectation(expectation_configuration=expectation_configuration)
{% endif %}
