class MyCustomNonCoreGeClass:
    """
    This class is used only for testing.
    E.g. ensuring appropriate usage stats messaging when using plugin functionality when the custom class is not a core GX class type.
    """

    def self_check(self, pretty_print):
        pass
