from great_expectations.data_context.store import ExpectationsStore


class MyCustomExpectationsStore(ExpectationsStore):
    """
    This class is used only for testing.
    E.g. ensuring appropriate usage stats messaging when using plugin functionality.
    """

    pass
